from __future__ import annotations

from matplotlib.axes import Axes

from maidr.core.enum import PlotType
from maidr.core.plot.maidr_plot import MaidrPlot
from maidr.core.plot.barplot import BarPlot
from maidr.core.plot.boxplot import BoxPlot
from maidr.core.plot.heatmap import HeatPlot
from maidr.core.plot.histogram import HistPlot
from maidr.core.plot.lineplot import LinePlot
from maidr.core.plot.scatterplot import ScatterPlot
from maidr.core.plot.grouped_barplot import GroupedBarPlot


class MaidrPlotFactory:
    """
    A factory for creating instances of ``maidr.core.maidr.MaidrPlot`` based on the
    plot type.

    Warnings
    --------
    End users will typically not have to use this class directly.

    See Also
    --------
    MaidrPlot : The base class for MAIDR plot data objects.
    PlotType : An enumeration of types of plots supported within MAIDR.
    """

    @staticmethod
    def create(ax: Axes, plot_type: PlotType, **kwargs) -> MaidrPlot:
        if PlotType.BAR == plot_type or PlotType.COUNT == plot_type:
            return BarPlot(ax)
        elif PlotType.BOX == plot_type:
            return BoxPlot(ax, **kwargs)
        elif PlotType.HEAT == plot_type:
            return HeatPlot(ax, **kwargs)
        elif PlotType.HIST == plot_type:
            return HistPlot(ax)
        elif PlotType.LINE == plot_type:
            return LinePlot(ax)
        elif PlotType.SCATTER == plot_type:
            return ScatterPlot(ax)
        elif PlotType.DODGED == plot_type or PlotType.STACKED == plot_type:
            return GroupedBarPlot(ax, plot_type)
        else:
            raise TypeError(f"Unsupported plot type: {plot_type}.")



from __future__ import annotations

from matplotlib.axes import Axes
from matplotlib.container import BarContainer

from maidr.core.enum import MaidrKey, PlotType
from maidr.core.plot import MaidrPlot
from maidr.exception import ExtractionError
from maidr.util.mixin import (
    ContainerExtractorMixin,
    DictMergerMixin,
    LevelExtractorMixin,
)


class GroupedBarPlot(
    MaidrPlot, ContainerExtractorMixin, LevelExtractorMixin, DictMergerMixin
):
    def __init__(self, ax: Axes, plot_type: PlotType) -> None:
        super().__init__(ax, plot_type)
        self._support_highlighting = False

    def _extract_axes_data(self) -> dict:
        base_ax_schema = super()._extract_axes_data()
        grouped_ax_schema = {
            MaidrKey.X.value: {
                MaidrKey.LEVEL.value: self.extract_level(self.ax, MaidrKey.X),
            },
            MaidrKey.FILL.value: {
                MaidrKey.LABEL.value: "Fill",
                MaidrKey.LEVEL.value: self.extract_level(self.ax, MaidrKey.FILL),
            },
        }
        return self.merge_dict(base_ax_schema, grouped_ax_schema)

    def _extract_plot_data(self):
        plot = self.extract_container(self.ax, BarContainer, include_all=True)
        data = self._extract_grouped_bar_data(plot)

        if data is None:
            raise ExtractionError(self.type, plot)

        return data

    def _extract_grouped_bar_data(
        self, plot: list[BarContainer] | None
    ) -> list[dict] | None:
        if plot is None:
            return None

        x_level = self.extract_level(self.ax)
        data = []

        for container in plot:
            if len(x_level) != len(container.patches):
                return None

            for x, y in zip(x_level, container.patches):
                data.append(
                    {
                        MaidrKey.X.value: x,
                        MaidrKey.FILL.value: container.get_label(),
                        MaidrKey.Y.value: float(y.get_height()),
                    }
                )

        return data



from __future__ import annotations

import wrapt

from matplotlib.axes import Axes
from matplotlib.lines import Line2D

from maidr.core.enum import PlotType
from maidr.patch.common import common


def line(wrapped, instance, args, kwargs) -> Axes | list[Line2D]:
    return common(PlotType.LINE, wrapped, instance, args, kwargs)


# Patch matplotlib function.
wrapt.wrap_function_wrapper(Axes, "plot", line)

# Patch seaborn function.
wrapt.wrap_function_wrapper("seaborn", "lineplot", line)



import json
import os


class Environment:
    @staticmethod
    def is_interactive_shell() -> bool:
        """Return True if the environment is an interactive shell."""
        try:
            from IPython.core.interactiveshell import InteractiveShell

            return (
                InteractiveShell.initialized()
                and InteractiveShell.instance() is not None
            )
        except ImportError:
            return False

    @staticmethod
    def is_notebook() -> bool:
        """Return True if the environment is a Jupyter notebook."""
        try:
            from IPython import get_ipython  # type: ignore

            return get_ipython() is not None and (
                "ipykernel" in str(get_ipython())
                or "google.colab" in str(get_ipython())
            )
        except ImportError:
            return False

    @staticmethod
    def get_renderer() -> str:
        """Return renderer which can be ipython or browser."""
        try:
            import IPython  # pyright: ignore[reportUnknownVariableType]

            ipy = (  # pyright: ignore[reportUnknownVariableType]
                IPython.get_ipython()  # pyright: ignore[reportUnknownMemberType, reportPrivateImportUsage]
            )
            renderer = "ipython" if ipy else "browser"
        except ImportError:
            renderer = "browser"
        return renderer

    @staticmethod
    def initialize_llm_secrets(unique_id: str) -> str:
        """Inject the LLM API keys into the MAIDR instance."""

        gemini_api_key = os.getenv("GEMINI_API_KEY")
        openai_api_key = os.getenv("OPENAI_API_KEY")

        # Default settings for the MAIDR instance
        settings = {
            "vol": "0.5",
            "autoPlayRate": "500",
            "brailleDisplayLength": "32",
            "colorSelected": "#03c809",
            "MIN_FREQUENCY": "200",
            "MAX_FREQUENCY": "1000",
            "keypressInterval": "2000",
            "ariaMode": "assertive",
            "openAIAuthKey": "",
            "geminiAuthKey": "",
            "skillLevel": "basic",
            "skillLevelOther": "",
            "LLMModel": "openai",
            "LLMPreferences": "",
            "LLMOpenAiMulti": False,
            "LLMGeminiMulti": False,
            "autoInitLLM": True,
        }

        if gemini_api_key is not None and openai_api_key is not None:
            settings["geminiAuthKey"] = gemini_api_key
            settings["openAIAuthKey"] = openai_api_key
            settings["LLMOpenAiMulti"] = True
            settings["LLMGeminiMulti"] = True
            settings["LLMModel"] = "multi"
        elif openai_api_key is not None:
            settings["LLMOpenAiMulti"] = True
            settings["openAIAuthKey"] = openai_api_key
            settings["LLMModel"] = "openai"
        elif gemini_api_key is not None:
            settings["LLMGeminiMulti"] = True
            settings["geminiAuthKey"] = gemini_api_key
            settings["LLMModel"] = "gemini"

        settings_data = json.dumps(settings)

        keys_injection_script = f"""
            function addKeyValueLocalStorage(iframeId, key, value) {{
                const iframe = document.getElementById(iframeId);
                if (iframe && iframe.contentWindow) {{
                    try {{
                        iframe.contentWindow.localStorage.setItem(key, value);
                    }} catch (error) {{
                        console.error('Error accessing iframe localStorage:', error);
                    }}
                }} else {{
                    console.error('Iframe not found or inaccessible.');
                }}
            }}
            addKeyValueLocalStorage(
                '{unique_id}', 'settings_data', JSON.stringify({settings_data})
            );
        """

        return keys_injection_script



from __future__ import annotations

from matplotlib.axes import Axes
from matplotlib.container import BarContainer

from maidr.core.enum import MaidrKey, PlotType
from maidr.core.plot import MaidrPlot
from maidr.exception import ExtractionError
from maidr.util.mixin import (
    ContainerExtractorMixin,
    DictMergerMixin,
    LevelExtractorMixin,
)


class BarPlot(MaidrPlot, ContainerExtractorMixin, LevelExtractorMixin, DictMergerMixin):
    def __init__(self, ax: Axes) -> None:
        super().__init__(ax, PlotType.BAR)

    def _extract_axes_data(self) -> dict:
        base_schema = super()._extract_axes_data()
        bar_ax_schema = {
            MaidrKey.X: {
                MaidrKey.LEVEL: self.extract_level(self.ax),
            },
        }
        return self.merge_dict(base_schema, bar_ax_schema)

    def _extract_plot_data(self) -> list:
        plot = self.extract_container(self.ax, BarContainer, include_all=True)
        data = self._extract_bar_container_data(plot)

        if data is None:
            raise ExtractionError(self.type, plot)

        return data

    def _extract_bar_container_data(
        self, plot: list[BarContainer] | None
    ) -> list | None:
        if plot is None:
            return None

        # Since v0.13, Seaborn has transitioned from using `list[Patch]` to
        # `list[BarContainers] for plotting bar plots.
        # So, extract data correspondingly based on the level.
        # Flatten all the `list[BarContainer]` to `list[Patch]`.
        plot = [patch for container in plot for patch in container.patches]
        level = self.extract_level(self.ax)

        if len(plot) != len(level):
            return None

        # Tag the elements for highlighting.
        self._elements.extend(plot)

        return [float(patch.get_height()) for patch in plot]



from __future__ import annotations

from matplotlib.axes import Axes
from matplotlib.container import BarContainer

from maidr.core.enum import MaidrKey, PlotType
from maidr.core.plot import MaidrPlot
from maidr.exception import ExtractionError
from maidr.util.mixin import ContainerExtractorMixin


class HistPlot(MaidrPlot, ContainerExtractorMixin):
    def __init__(self, ax: Axes) -> None:
        super().__init__(ax, PlotType.HIST)

    def _extract_plot_data(self) -> list[dict]:
        plot = self.extract_container(self.ax, BarContainer)
        data = self._extract_bar_container_data(plot)

        if data is None:
            raise ExtractionError(self.type, plot)

        return data

    def _extract_bar_container_data(
        self, plot: BarContainer | None
    ) -> list[dict] | None:
        if plot is None or plot.patches is None:
            return None

        data = []
        for patch in plot.patches:
            y = float(patch.get_height())
            x = float(patch.get_x())
            width = float(patch.get_width())

            data.append(
                {
                    MaidrKey.Y.value: y,
                    MaidrKey.X.value: x + width / 2,
                    MaidrKey.X_MIN.value: x,
                    MaidrKey.X_MAX.value: x + width,
                    MaidrKey.Y_MIN.value: 0,
                    MaidrKey.Y_MAX.value: y,
                }
            )

        # Tag the elements for highlighting
        self._elements.extend(plot.patches)

        return data



from __future__ import annotations

import contextlib
import contextvars
import threading

import wrapt

from maidr.core.plot.boxplot import BoxPlotContainer


class ContextManager:
    _instance = None
    _lock = threading.Lock()

    _internal_context = contextvars.ContextVar("internal_context", default=False)

    def __new__(cls):
        if not cls._instance:
            with cls._lock:
                if not cls._instance:
                    cls._instance = super(ContextManager, cls).__new__()
        return cls._instance

    @classmethod
    def is_internal_context(cls):
        return cls._internal_context.get()

    @classmethod
    @contextlib.contextmanager
    def set_internal_context(cls):
        token_internal_context = cls._internal_context.set(True)
        try:
            yield
        finally:
            cls._internal_context.reset(token_internal_context)


@wrapt.decorator
def manage_context(wrapped=None, _=None, args=None, kwargs=None):
    # Don't proceed if the call is made internally by the patched function.
    if ContextManager.is_internal_context():
        return wrapped(*args, **kwargs)

    # Set the internal context to avoid cyclic processing.
    with ContextManager.set_internal_context():
        return wrapped(*args, **kwargs)


class BoxplotContextManager(ContextManager):
    _bxp_context = contextvars.ContextVar("bxp_context", default=BoxPlotContainer())

    @classmethod
    @contextlib.contextmanager
    def set_internal_context(cls):
        with super(BoxplotContextManager, cls).set_internal_context():
            token = cls._bxp_context.set(BoxPlotContainer())
            try:
                yield cls.get_bxp_context()
            finally:
                cls._bxp_context.reset(token)

    @classmethod
    def get_bxp_context(cls) -> BoxPlotContainer:
        return cls._bxp_context.get()

    @classmethod
    def add_bxp_context(cls, bxp_context: dict) -> None:
        cls.get_bxp_context().add_artists(bxp_context)

    @classmethod
    def set_bxp_orientation(cls, orientation: str) -> None:
        cls.get_bxp_context().set_orientation(orientation)


class HighlightContextManager:
    _instance = None
    _lock = threading.Lock()

    _maidr_element = contextvars.ContextVar("_maidr_element", default=False)
    _elements = contextvars.ContextVar("elements", default=[])

    def __new__(cls):
        if not cls._instance:
            with cls._lock:
                if not cls._instance:
                    cls._instance = super(HighlightContextManager, cls).__new__()
        return cls._instance

    @classmethod
    def is_maidr_element(cls):
        return cls._maidr_element.get()

    @classmethod
    @contextlib.contextmanager
    def set_maidr_element(cls, element):
        if element not in cls._elements.get():
            yield
            return

        token_maidr_element = cls._maidr_element.set(True)
        try:
            yield
        finally:
            cls._maidr_element.reset(token_maidr_element)
            # Remove element from the context list after tagging
            new_elements = cls._elements.get().copy()
            new_elements.remove(element)
            cls._elements.set(new_elements)

    @classmethod
    @contextlib.contextmanager
    def set_maidr_elements(cls, elements: list):
        token_paths = cls._elements.set(elements)
        try:
            yield
        finally:
            cls._elements.reset(token_paths)



from __future__ import annotations

import wrapt

from matplotlib.axes import Axes
from matplotlib.collections import PathCollection

from maidr.core.enum import PlotType
from maidr.patch.common import common


def scatter(wrapped, instance, args, kwargs) -> Axes | PathCollection:
    return common(PlotType.SCATTER, wrapped, instance, args, kwargs)


# Patch matplotlib function.
wrapt.wrap_function_wrapper(Axes, "scatter", scatter)

# Patch seaborn function.
wrapt.wrap_function_wrapper("seaborn", "scatterplot", scatter)



from __future__ import annotations

import numpy.ma as ma

from matplotlib.axes import Axes
from matplotlib.collections import PathCollection

from maidr.core.enum import MaidrKey, PlotType
from maidr.core.plot import MaidrPlot
from maidr.exception import ExtractionError
from maidr.util.mixin import CollectionExtractorMixin


class ScatterPlot(MaidrPlot, CollectionExtractorMixin):
    def __init__(self, ax: Axes) -> None:
        super().__init__(ax, PlotType.SCATTER)

    def _get_selector(self) -> str:
        return "g[maidr='true'] > use"

    def _extract_plot_data(self) -> list[dict]:
        plot = self.extract_collection(self.ax, PathCollection)
        data = self._extract_point_data(plot)

        if data is None:
            raise ExtractionError(self.type, plot)

        return data

    def _extract_point_data(self, plot: PathCollection | None) -> list[dict] | None:
        if plot is None or plot.get_offsets() is None:
            return None

        # Tag the elements for highlighting.
        self._elements.append(plot)

        return [
            {
                MaidrKey.X: float(x),
                MaidrKey.Y: float(y),
            }
            for x, y in ma.getdata(plot.get_offsets())
        ]
